# ruff: noqa: I001
from .api import (  # noqa
    bluelet,
)

from .core import (  # noqa
    BlueletCoro as Coro,
    BlueletExcInfo as ExcInfo,
    CoreBlueletEvent as CoreEvent,
    DelegationBlueletEvent as DelegationEvent,
    ExceptionBlueletEvent as ExceptionEvent,
    JoinBlueletEvent as JoinEvent,
    KillBlueletEvent as KillEvent,
    ReturnBlueletEvent as ReturnEvent,
    SleepBlueletEvent as SleepEvent,
    SpawnBlueletEvent as SpawnEvent,
    ValueBlueletEvent as ValueEvent,
)

from .events import (  # noqa
    BlueletEvent as Event,
    BlueletFuture as Future,
    BlueletHasFileno as HasFileno,
    BlueletWaitable as Waitable,
    BlueletWaitables as Waitables,
    WaitableBlueletEvent as WaitableEvent,
)

from .files import (  # noqa
    FileBlueletEvent as FileEvent,
    ReadBlueletEvent as ReadEvent,
    WriteBlueletEvent as WriteEvent,
)

from .runner import (  # noqa
    BlueletCoroException as CoroException,
)

from .sockets import (  # noqa
    AcceptBlueletEvent as AcceptEvent,
    BlueletConnection as Connection,
    BlueletListener as Listener,
    ReceiveBlueletEvent as ReceiveEvent,
    SendBlueletEvent as SendEvent,
    SocketBlueletEvent as SocketEvent,
    SocketClosedBlueletError as SocketClosedError,
)


##


call = bluelet.call
end = bluelet.end
join = bluelet.join
kill = bluelet.kill
null = bluelet.null
sleep = bluelet.sleep
spawn = bluelet.spawn

read = bluelet.read
write = bluelet.write

run = bluelet.run

connect = bluelet.connect
server = bluelet.server



"""
FIXME:
 - refuse to marshal anon params
"""
import enum
import typing as ta

from ... import cached
from ... import check
from ... import collections as col
from ... import dataclasses as dc
from ... import lang
from ... import marshal as msh
from .base import Node
from .binary import BinaryOp
from .binary import BinaryOps
from .exprs import Expr
from .inserts import Values
from .multi import MultiKind
from .relations import Relation
from .selects import Select
from .stmts import Stmt
from .unary import UnaryOp
from .unary import UnaryOps


@dc.dataclass(frozen=True)
class OpMarshalerUnmarshaler(msh.Marshaler, msh.Unmarshaler):
    ty: type
    ns: type[lang.Namespace]

    @cached.property
    @dc.init
    def by_name(self) -> ta.Mapping[str, ta.Any]:
        return col.make_map(((o.name, o) for _, o in self.ns), strict=True)

    def marshal(self, ctx: msh.MarshalContext, o: ta.Any) -> msh.Value:
        return check.isinstance(o, self.ty).name  # type: ignore  # noqa

    def unmarshal(self, ctx: msh.UnmarshalContext, v: msh.Value) -> ta.Any:
        return self.by_name[check.isinstance(v, str)]


@dc.dataclass(frozen=True)
class LowerEnumMarshaler(msh.Marshaler, msh.Unmarshaler):
    ty: type[enum.Enum]

    @cached.property
    @dc.init
    def by_name(self) -> ta.Mapping[str, ta.Any]:
        return col.make_map(((o.name.lower(), o) for o in self.ty), strict=True)

    def marshal(self, ctx: msh.MarshalContext, o: ta.Any) -> msh.Value:
        return o.name.lower()

    def unmarshal(self, ctx: msh.UnmarshalContext, v: msh.Value) -> ta.Any:
        return self.by_name[check.isinstance(v, str).lower()]


@lang.static_init
def _install_standard_marshalling() -> None:
    for ty, ns in [
        (BinaryOp, BinaryOps),
        (UnaryOp, UnaryOps),
    ]:
        msh.STANDARD_MARSHALER_FACTORIES[0:0] = [msh.TypeMapMarshalerFactory({ty: OpMarshalerUnmarshaler(ty, ns)})]
        msh.STANDARD_UNMARSHALER_FACTORIES[0:0] = [msh.TypeMapUnmarshalerFactory({ty: OpMarshalerUnmarshaler(ty, ns)})]

    msh.STANDARD_MARSHALER_FACTORIES[0:0] = [msh.TypeMapMarshalerFactory({MultiKind: LowerEnumMarshaler(MultiKind)})]
    msh.STANDARD_UNMARSHALER_FACTORIES[0:0] = [msh.TypeMapUnmarshalerFactory({MultiKind: LowerEnumMarshaler(MultiKind)})]  # noqa

    for cls in [
        Expr,
        Node,
        Relation,
        Stmt,
    ]:
        p = msh.polymorphism_from_subclasses(cls, naming=msh.Naming.SNAKE)
        msh.STANDARD_MARSHALER_FACTORIES[0:0] = [msh.PolymorphismMarshalerFactory(p)]
        msh.STANDARD_UNMARSHALER_FACTORIES[0:0] = [msh.PolymorphismUnmarshalerFactory(p)]

    insert_data_impls = msh.Impls([
        msh.Impl(Values, 'values'),
        msh.Impl(Select, 'select'),
    ])
    msh.STANDARD_MARSHALER_FACTORIES[0:0] = [msh.PolymorphismUnionMarshalerFactory(insert_data_impls)]
    msh.STANDARD_UNMARSHALER_FACTORIES[0:0] = [msh.PolymorphismUnionUnmarshalerFactory(insert_data_impls)]



from .base import Backend
from .orjson import ORJSON_BACKEND
from .std import STD_BACKEND
from .ujson import UJSON_BACKEND


DEFAULT_BACKED: Backend
if ORJSON_BACKEND is not None:
    DEFAULT_BACKED = ORJSON_BACKEND
elif UJSON_BACKEND is not None:
    DEFAULT_BACKED = UJSON_BACKEND
else:
    DEFAULT_BACKED = STD_BACKEND



import sys


REQUIRED_PYTHON_VERSION = (3, 12)


def check_runtime_version() -> None:
    if sys.version_info < REQUIRED_PYTHON_VERSION:
        raise OSError(
            f'Requires python {REQUIRED_PYTHON_VERSION}, got {sys.version_info} from {sys.executable}')  # noqa


def is_gil_enabled() -> bool:
    if (fn := getattr(sys, '_is_gil_enabled', None)) is not None:
        return fn()
    return True



# ruff: noqa: UP006 UP007
# @omlish-lite
"""
TODO:
 - env vars
 - coalescing - {$FOO|$BAR|baz}
 - minja?
"""
import typing as ta

from .rewriting import ConfigRewriter


T = ta.TypeVar('T')
StrT = ta.TypeVar('StrT', bound=str)


##


class StringConfigRewriter(ConfigRewriter):
    def __init__(
            self,
            fn: ta.Optional[ta.Callable[[str], str]] = None,
    ) -> None:
        super().__init__()

        self._fn = fn

    def rewrite_str(self, ctx: ConfigRewriter.Context[StrT]) -> StrT:
        v = ctx.obj
        if (fn := self._fn) is not None:
            if not ctx.raw:
                v = fn(v)  # type: ignore
        return type(v)(v)


##


def format_config_strings(v: T, rpl: ta.Mapping[str, str]) -> T:
    def fn(v):
        return v.format(**rpl)

    return StringConfigRewriter(fn)(v)



import abc
import typing as ta
import weakref

from .dispatch import find_impl


T = ta.TypeVar('T')


##


class Dispatcher(ta.Generic[T]):
    def __init__(self) -> None:
        super().__init__()

        self._impls_by_arg_cls: dict[type, T] = {}
        self._dispatch_cache: dict[ta.Any, T | None] = {}

        def cache_remove(k, self_ref=weakref.ref(self)):
            if (ref_self := self_ref()) is not None:
                cache = ref_self._dispatch_cache  # noqa
                try:
                    del cache[k]
                except KeyError:
                    pass

        self._cache_remove = cache_remove

        self._cache_token: ta.Any = None

    def cache_size(self) -> int:
        return len(self._dispatch_cache)

    def register(self, impl: T, cls_col: ta.Iterable[type]) -> T:
        for cls in cls_col:
            self._impls_by_arg_cls[cls] = impl

            if self._cache_token is None and hasattr(cls, '__abstractmethods__'):
                self._cache_token = abc.get_cache_token()

        self._dispatch_cache.clear()
        return impl

    def dispatch(self, cls: type) -> T | None:
        if self._cache_token is not None and (current_token := abc.get_cache_token()) != self._cache_token:
            self._dispatch_cache.clear()
            self._cache_token = current_token

        cls_ref = weakref.ref(cls)
        try:
            return self._dispatch_cache[cls_ref]
        except KeyError:
            pass
        del cls_ref

        impl: T | None
        try:
            impl = self._impls_by_arg_cls[cls]
        except KeyError:
            impl = find_impl(cls, self._impls_by_arg_cls)

        self._dispatch_cache[weakref.ref(cls, self._cache_remove)] = impl
        return impl



import dataclasses as dc
import functools
import time
import typing as ta


F = ta.TypeVar('F')
T = ta.TypeVar('T')
P = ta.ParamSpec('P')
CallableT = ta.TypeVar('CallableT', bound=ta.Callable)


##


def is_lambda(f: ta.Any) -> bool:
    l = lambda: 0
    return isinstance(f, type(l)) and f.__name__ == l.__name__


##


def maybe_call(obj: ta.Any, att: str, *args, default: ta.Any = None, **kwargs) -> ta.Any:
    try:
        fn = getattr(obj, att)
    except AttributeError:
        return default
    else:
        return fn(*args, **kwargs)


def recurse(fn: ta.Callable[..., T], *args, **kwargs) -> T:
    def rec(*args, **kwargs) -> T:
        return fn(rec, *args, **kwargs)

    return rec(*args, **kwargs)


##


def raise_(o: BaseException) -> ta.NoReturn:
    raise o


def raising(o: BaseException) -> ta.Callable[..., ta.NoReturn]:
    def inner(*args, **kwargs):
        raise o
    return inner


def try_(
        fn: ta.Callable[P, T],
        exc: type[Exception] | ta.Iterable[type[Exception]] = Exception,
        default: T | None = None,
) -> ta.Callable[P, T]:
    def inner(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except exct:
            return default

    exct = (exc,) if isinstance(exc, type) else tuple(exc)
    return inner


def finally_(fn: ta.Callable[P, T], fin: ta.Callable) -> ta.Callable[P, T]:
    def inner(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        finally:
            fin()

    return inner


##


def identity(obj: T) -> T:
    return obj


def opt_fn(fn: ta.Callable[[F], T]) -> ta.Callable[[F | None], T | None]:
    @functools.wraps(fn)
    def inner(v: F | None) -> T | None:
        if v is not None:
            return fn(v)
        else:
            return None
    return inner


class constant(ta.Generic[T]):  # noqa
    def __init__(self, obj: T) -> None:
        super().__init__()

        self._obj = obj

    def __call__(self) -> T:
        return self._obj


def is_none(o: ta.Any) -> bool:
    return o is None


def is_not_none(o: ta.Any) -> bool:
    return o is not None


def isinstance_of(class_or_tuple: ta.Any) -> ta.Callable[[ta.Any], bool]:
    return lambda o: isinstance(o, class_or_tuple)


def issubclass_of(class_or_tuple: ta.Any) -> ta.Callable[[ta.Any], bool]:
    return lambda o: issubclass(o, class_or_tuple)


##


class VoidError(Exception):
    pass


class Void:
    def __new__(cls, *args: ta.Any, **kwargs: ta.Any) -> None:  # type: ignore  # noqa
        raise VoidError

    def __init_subclass__(cls, **kwargs: ta.Any) -> None:
        raise VoidError


def void(*args: ta.Any, **kwargs: ta.Any) -> ta.NoReturn:
    raise VoidError


##


def as_async(fn: ta.Callable[P, T]) -> ta.Callable[P, ta.Awaitable[T]]:
    @functools.wraps(fn)
    async def inner(*args, **kwargs):
        return fn(*args, **kwargs)
    return inner


##


_MISSING = object()


def periodically(
        fn: CallableT,
        interval_s: float,
        initial: ta.Any = _MISSING,
        *,
        include_runtime: bool = False,
) -> CallableT:
    nxt = time.time() + interval_s
    ret = initial

    @functools.wraps(fn)
    def inner(*args, **kwargs):
        nonlocal nxt, ret
        if time.time() >= nxt or ret is _MISSING:
            if include_runtime:
                nxt = time.time() + interval_s
            ret = fn(*args, **kwargs)
            if not include_runtime:
                nxt = time.time() + interval_s
        return ret

    return inner  # type: ignore


##


@dc.dataclass(init=False)
class Args:
    args: ta.Sequence[ta.Any]
    kwargs: ta.Mapping[str, ta.Any]

    def __init__(self, *args: ta.Any, **kwargs: ta.Any) -> None:
        super().__init__()
        self.args = args
        self.kwargs = kwargs

    def __call__(self, fn: ta.Callable[..., T]) -> T:
        return fn(*self.args, **self.kwargs)


##


def coalesce(*vs: T | None) -> T:
    for v in vs:
        if v is not None:
            return v
    raise ValueError('No value given')


def opt_coalesce(*vs: T | None) -> T | None:
    for v in vs:
        if v is not None:
            return v
    return None



"""
~> https://github.com/pallets/werkzeug/blob/7a76170c473c26685bdfa2774d083ba2386fc60f/src/werkzeug/security.py
"""
# Copyright 2007 Pallets
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1.  Redistributions of source code must retain the above copyright notice, this list of conditions and the following
#     disclaimer.
#
# 2.  Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
#     following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3.  Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
#     products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import hashlib
import hmac
import secrets


SALT_CHARS = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'
DEFAULT_PBKDF2_ITERATIONS = 600_000


def gen_salt(length: int) -> str:
    if length <= 0:
        raise ValueError('Salt length must be at least 1.')

    return ''.join(secrets.choice(SALT_CHARS) for _ in range(length))


def _hash_internal(
        method: str,
        salt: str,
        password: str,
) -> tuple[str, str]:
    method, *args = method.split(':')
    salt_bytes = salt.encode()
    password_bytes = password.encode()

    if method == 'scrypt':
        if not args:
            n = 2 ** 15
            r = 8
            p = 1
        else:
            try:
                n, r, p = map(int, args)
            except ValueError:
                raise ValueError("'scrypt' takes 3 arguments.") from None

        maxmem = 132 * n * r * p  # ideally 128, but some extra seems needed

        return (
            hashlib.scrypt(
                password_bytes,
                salt=salt_bytes,
                n=n,
                r=r,
                p=p,
                maxmem=maxmem,
            ).hex(),
            f'scrypt:{n}:{r}:{p}',
        )

    elif method == 'pbkdf2':
        len_args = len(args)

        if len_args == 0:
            hash_name = 'sha256'
            iterations = DEFAULT_PBKDF2_ITERATIONS
        elif len_args == 1:
            hash_name = args[0]
            iterations = DEFAULT_PBKDF2_ITERATIONS
        elif len_args == 2:
            hash_name = args[0]
            iterations = int(args[1])
        else:
            raise ValueError("'pbkdf2' takes 2 arguments.")

        return (
            hashlib.pbkdf2_hmac(
                hash_name,
                password_bytes,
                salt_bytes,
                iterations,
            ).hex(),
            f'pbkdf2:{hash_name}:{iterations}',
        )

    else:
        raise ValueError(f"Invalid hash method '{method}'.")


def generate_password_hash(
        password: str,
        method: str = 'scrypt',
        salt_length: int = 16,
) -> str:
    salt = gen_salt(salt_length)
    h, actual_method = _hash_internal(method, salt, password)
    return f'{actual_method}${salt}${h}'


def check_password_hash(pwhash: str, password: str) -> bool:
    try:
        method, salt, hashval = pwhash.split('$', 2)
    except ValueError:
        return False

    return hmac.compare_digest(_hash_internal(method, salt, password)[0], hashval)



# ruff: noqa: UP006 UP007
# @omlish-lite
import abc
import contextlib
import logging
import os
import shlex
import subprocess
import sys
import time
import typing as ta

from .lite.runtime import is_debugger_attached


T = ta.TypeVar('T')
SubprocessChannelOption = ta.Literal['pipe', 'stdout', 'devnull']  # ta.TypeAlias


##


SUBPROCESS_CHANNEL_OPTION_VALUES: ta.Mapping[SubprocessChannelOption, int] = {
    'pipe': subprocess.PIPE,
    'stdout': subprocess.STDOUT,
    'devnull': subprocess.DEVNULL,
}


##


_SUBPROCESS_SHELL_WRAP_EXECS = False


def subprocess_shell_wrap_exec(*cmd: str) -> ta.Tuple[str, ...]:
    return ('sh', '-c', ' '.join(map(shlex.quote, cmd)))


def subprocess_maybe_shell_wrap_exec(*cmd: str) -> ta.Tuple[str, ...]:
    if _SUBPROCESS_SHELL_WRAP_EXECS or is_debugger_attached():
        return subprocess_shell_wrap_exec(*cmd)
    else:
        return cmd


##


def subprocess_close(
        proc: subprocess.Popen,
        timeout: ta.Optional[float] = None,
) -> None:
    # TODO: terminate, sleep, kill
    if proc.stdout:
        proc.stdout.close()
    if proc.stderr:
        proc.stderr.close()
    if proc.stdin:
        proc.stdin.close()

    proc.wait(timeout)


##


class BaseSubprocesses(abc.ABC):  # noqa
    DEFAULT_LOGGER: ta.ClassVar[ta.Optional[logging.Logger]] = None

    def __init__(
            self,
            *,
            log: ta.Optional[logging.Logger] = None,
            try_exceptions: ta.Optional[ta.Tuple[ta.Type[Exception], ...]] = None,
    ) -> None:
        super().__init__()

        self._log = log if log is not None else self.DEFAULT_LOGGER
        self._try_exceptions = try_exceptions if try_exceptions is not None else self.DEFAULT_TRY_EXCEPTIONS

    def set_logger(self, log: ta.Optional[logging.Logger]) -> None:
        self._log = log

    #

    def prepare_args(
            self,
            *cmd: str,
            env: ta.Optional[ta.Mapping[str, ta.Any]] = None,
            extra_env: ta.Optional[ta.Mapping[str, ta.Any]] = None,
            quiet: bool = False,
            shell: bool = False,
            **kwargs: ta.Any,
    ) -> ta.Tuple[ta.Tuple[ta.Any, ...], ta.Dict[str, ta.Any]]:
        if self._log:
            self._log.debug('Subprocesses.prepare_args: cmd=%r', cmd)
            if extra_env:
                self._log.debug('Subprocesses.prepare_args: extra_env=%r', extra_env)

        if extra_env:
            env = {**(env if env is not None else os.environ), **extra_env}

        if quiet and 'stderr' not in kwargs:
            if self._log and not self._log.isEnabledFor(logging.DEBUG):
                kwargs['stderr'] = subprocess.DEVNULL

        if not shell:
            cmd = subprocess_maybe_shell_wrap_exec(*cmd)

        return cmd, dict(
            env=env,
            shell=shell,
            **kwargs,
        )

    @contextlib.contextmanager
    def wrap_call(self, *cmd: ta.Any, **kwargs: ta.Any) -> ta.Iterator[None]:
        start_time = time.time()
        try:
            if self._log:
                self._log.debug('Subprocesses.wrap_call.try: cmd=%r', cmd)
            yield

        except Exception as exc:  # noqa
            if self._log:
                self._log.debug('Subprocesses.wrap_call.except: exc=%r', exc)
            raise

        finally:
            end_time = time.time()
            elapsed_s = end_time - start_time
            if self._log:
                self._log.debug('sSubprocesses.wrap_call.finally: elapsed_s=%f cmd=%r', elapsed_s, cmd)

    @contextlib.contextmanager
    def prepare_and_wrap(
            self,
            *cmd: ta.Any,
            **kwargs: ta.Any,
    ) -> ta.Iterator[ta.Tuple[
        ta.Tuple[ta.Any, ...],
        ta.Dict[str, ta.Any],
    ]]:
        cmd, kwargs = self.prepare_args(*cmd, **kwargs)
        with self.wrap_call(*cmd, **kwargs):
            yield cmd, kwargs

    #

    DEFAULT_TRY_EXCEPTIONS: ta.Tuple[ta.Type[Exception], ...] = (
        FileNotFoundError,
        subprocess.CalledProcessError,
    )

    def try_fn(
            self,
            fn: ta.Callable[..., T],
            *cmd: str,
            try_exceptions: ta.Optional[ta.Tuple[ta.Type[Exception], ...]] = None,
            **kwargs: ta.Any,
    ) -> ta.Union[T, Exception]:
        if try_exceptions is None:
            try_exceptions = self._try_exceptions

        try:
            return fn(*cmd, **kwargs)

        except try_exceptions as e:  # noqa
            if self._log and self._log.isEnabledFor(logging.DEBUG):
                self._log.exception('command failed')
            return e

    async def async_try_fn(
            self,
            fn: ta.Callable[..., ta.Awaitable[T]],
            *cmd: ta.Any,
            try_exceptions: ta.Optional[ta.Tuple[ta.Type[Exception], ...]] = None,
            **kwargs: ta.Any,
    ) -> ta.Union[T, Exception]:
        if try_exceptions is None:
            try_exceptions = self._try_exceptions

        try:
            return await fn(*cmd, **kwargs)

        except try_exceptions as e:  # noqa
            if self._log and self._log.isEnabledFor(logging.DEBUG):
                self._log.exception('command failed')
            return e


##


class AbstractSubprocesses(BaseSubprocesses, abc.ABC):
    @abc.abstractmethod
    def check_call(
            self,
            *cmd: str,
            stdout: ta.Any = sys.stderr,
            **kwargs: ta.Any,
    ) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def check_output(
            self,
            *cmd: str,
            **kwargs: ta.Any,
    ) -> bytes:
        raise NotImplementedError

    #

    def check_output_str(
            self,
            *cmd: str,
            **kwargs: ta.Any,
    ) -> str:
        return self.check_output(*cmd, **kwargs).decode().strip()

    #

    def try_call(
            self,
            *cmd: str,
            **kwargs: ta.Any,
    ) -> bool:
        if isinstance(self.try_fn(self.check_call, *cmd, **kwargs), Exception):
            return False
        else:
            return True

    def try_output(
            self,
            *cmd: str,
            **kwargs: ta.Any,
    ) -> ta.Optional[bytes]:
        if isinstance(ret := self.try_fn(self.check_output, *cmd, **kwargs), Exception):
            return None
        else:
            return ret

    def try_output_str(
            self,
            *cmd: str,
            **kwargs: ta.Any,
    ) -> ta.Optional[str]:
        if (ret := self.try_output(*cmd, **kwargs)) is None:
            return None
        else:
            return ret.decode().strip()


##


class Subprocesses(AbstractSubprocesses):
    def check_call(
            self,
            *cmd: str,
            stdout: ta.Any = sys.stderr,
            **kwargs: ta.Any,
    ) -> None:
        with self.prepare_and_wrap(*cmd, stdout=stdout, **kwargs) as (cmd, kwargs):  # noqa
            subprocess.check_call(cmd, **kwargs)

    def check_output(
            self,
            *cmd: str,
            **kwargs: ta.Any,
    ) -> bytes:
        with self.prepare_and_wrap(*cmd, **kwargs) as (cmd, kwargs):  # noqa
            return subprocess.check_output(cmd, **kwargs)


subprocesses = Subprocesses()


##


class AbstractAsyncSubprocesses(BaseSubprocesses):
    @abc.abstractmethod
    async def check_call(
            self,
            *cmd: str,
            stdout: ta.Any = sys.stderr,
            **kwargs: ta.Any,
    ) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    async def check_output(
            self,
            *cmd: str,
            **kwargs: ta.Any,
    ) -> bytes:
        raise NotImplementedError

    #

    async def check_output_str(
            self,
            *cmd: str,
            **kwargs: ta.Any,
    ) -> str:
        return (await self.check_output(*cmd, **kwargs)).decode().strip()

    #

    async def try_call(
            self,
            *cmd: str,
            **kwargs: ta.Any,
    ) -> bool:
        if isinstance(await self.async_try_fn(self.check_call, *cmd, **kwargs), Exception):
            return False
        else:
            return True

    async def try_output(
            self,
            *cmd: str,
            **kwargs: ta.Any,
    ) -> ta.Optional[bytes]:
        if isinstance(ret := await self.async_try_fn(self.check_output, *cmd, **kwargs), Exception):
            return None
        else:
            return ret

    async def try_output_str(
            self,
            *cmd: str,
            **kwargs: ta.Any,
    ) -> ta.Optional[str]:
        if (ret := await self.try_output(*cmd, **kwargs)) is None:
            return None
        else:
            return ret.decode().strip()



import contextlib
import os
import shutil
import tempfile
import typing as ta


@contextlib.contextmanager
def tmp_dir(
        root_dir: str | None = None,
        cleanup: bool = True,
        **kwargs: ta.Any,
) -> ta.Iterator[str]:
    path = tempfile.mkdtemp(dir=root_dir, **kwargs)
    try:
        yield path
    finally:
        if cleanup:
            shutil.rmtree(path, ignore_errors=True)


@contextlib.contextmanager
def tmp_file(
        root_dir: str | None = None,
        cleanup: bool = True,
        **kwargs: ta.Any,
) -> ta.Iterator[tempfile._TemporaryFileWrapper]:  # noqa
    with tempfile.NamedTemporaryFile(dir=root_dir, delete=False, **kwargs) as f:
        try:
            yield f
        finally:
            if cleanup:
                shutil.rmtree(f.name, ignore_errors=True)


def touch(self, mode: int = 0o666, exist_ok: bool = True) -> None:
    if exist_ok:
        # First try to bump modification time
        # Implementation note: GNU touch uses the UTIME_NOW option of the utimensat() / futimens() functions.
        try:
            os.utime(self, None)
        except OSError:
            pass
        else:
            return
    flags = os.O_CREAT | os.O_WRONLY
    if not exist_ok:
        flags |= os.O_EXCL
    fd = os.open(self, flags, mode)
    os.close(fd)



import typing as ta

from .. import lang


T = ta.TypeVar('T')
K = ta.TypeVar('K')
V = ta.TypeVar('V')


class Unmodifiable(lang.Abstract):
    pass


class UnmodifiableSequence(ta.Sequence[T], Unmodifiable, lang.Final):

    def __init__(self, target: ta.Sequence[T]) -> None:
        super().__init__()

        if target is None:
            raise TypeError(target)
        self._target = target

    def __repr__(self) -> str:
        return f'{type(self).__name__}({self._target!r})'

    def __contains__(self, x: ta.Any) -> bool:
        return x in self._target

    def __eq__(self, o: object) -> bool:
        return self._target == o

    def __ge__(self, other: ta.Any) -> bool:
        return self._target >= other

    def __getitem__(self, i: int | slice) -> T:  # type: ignore
        return self._target[i]  # type: ignore

    def __gt__(self, other: ta.Any) -> bool:
        return self._target > other

    def __iter__(self) -> ta.Iterator[T]:
        return iter(self._target)

    def __le__(self, other: ta.Any) -> bool:
        return self._target <= other

    def __len__(self) -> int:
        return len(self._target)

    def __lt__(self, other: ta.Any) -> bool:
        return self._target < other

    def __ne__(self, o: object) -> bool:
        return self._target != o

    def __reversed__(self) -> ta.Iterator[T]:
        return reversed(self._target)

    def count(self, x: ta.Any) -> int:
        return self._target.count(x)

    def index(self, x: ta.Any, *args, **kwargs) -> int:  # type: ignore
        return self._target.index(x, *args, **kwargs)


class UnmodifiableSet(ta.AbstractSet[T], Unmodifiable, lang.Final):

    def __init__(self, target: ta.AbstractSet[T]) -> None:
        super().__init__()

        if target is None:
            raise TypeError(target)
        self._target = target

    def __repr__(self) -> str:
        return f'{type(self).__name__}({self._target!r})'

    def __and__(self, s: ta.AbstractSet[ta.Any]) -> ta.AbstractSet[T]:
        return self._target & s

    def __contains__(self, x: ta.Any) -> bool:
        return x in self._target

    def __eq__(self, o: object) -> bool:
        return self._target == o

    def __ge__(self, s: ta.AbstractSet[ta.Any]) -> bool:
        return self._target >= s

    def __gt__(self, s: ta.AbstractSet[ta.Any]) -> bool:
        return self._target > s

    def __iter__(self) -> ta.Iterator[T]:
        return iter(self._target)

    def __le__(self, s: ta.AbstractSet[ta.Any]) -> bool:
        return self._target <= s

    def __len__(self) -> int:
        return len(self._target)

    def __lt__(self, s: ta.AbstractSet[ta.Any]) -> bool:
        return self._target > s

    def __ne__(self, o: object) -> bool:
        return self._target != o

    def __or__(self, s: ta.AbstractSet[T]) -> ta.AbstractSet[T]:  # type: ignore
        return self._target | s

    def __sub__(self, s: ta.AbstractSet[ta.Any]) -> ta.AbstractSet[T]:
        return self._target - s

    def __xor__(self, s: ta.AbstractSet[T]) -> ta.AbstractSet[T]:  # type: ignore
        return self._target ^ s

    def isdisjoint(self, s: ta.Iterable[ta.Any]) -> bool:
        return self._target.isdisjoint(s)


class UnmodifiableMapping(ta.Mapping[K, V], Unmodifiable, lang.Final):
    # Could be a MappingProxyType but isn't for consistency with the others.

    def __init__(self, target: ta.Mapping[K, V]) -> None:
        super().__init__()

        if target is None:
            raise TypeError(target)
        self._target = target

    def __repr__(self) -> str:
        return f'{type(self).__name__}({self._target!r})'

    def __contains__(self, o: ta.Any) -> bool:
        return o in self._target

    def __eq__(self, o: object) -> bool:
        return self._target == o

    def __ge__(self, other: ta.Any) -> bool:
        return self._target >= other

    def __getitem__(self, k: K) -> V:
        return self._target[k]

    def __gt__(self, other: ta.Any) -> bool:
        return self._target > other

    def __iter__(self) -> ta.Iterator[T]:
        return iter(self._target)  # type: ignore

    def __le__(self, other: ta.Any) -> bool:
        return self._target <= other

    def __len__(self) -> int:
        return len(self._target)

    def __lt__(self, other: ta.Any) -> bool:
        return self._target < other

    def __ne__(self, o: object) -> bool:
        return self._target != o

    def get(self, k: K, default=None) -> V | None:  # type: ignore
        return self._target.get(k)

    def items(self) -> ta.ItemsView[K, V]:
        return self._target.items()

    def keys(self) -> ta.KeysView[K]:
        return self._target.keys()

    def values(self) -> ta.ValuesView[V]:
        return self._target.values()



"""
load(fp: File) -> ta.Any | uj.JSONDecodeError
loads(s: str | bytes | bytearray) -> ta.Any | uj.JSONDecodeError
dump(obj: ta.Any, fp: File, **DumpOpts) -> None
dumps(obj: ta.Any, **DumpOpts) -> None
"""
import dataclasses as dc
import typing as ta

from .... import lang
from ..consts import PRETTY_INDENT
from .base import Backend


if ta.TYPE_CHECKING:
    import ujson as uj
else:
    uj = lang.proxy_import('ujson')


##


@dc.dataclass(frozen=True, kw_only=True)
class DumpOpts:
    """
    https://github.com/ultrajson/ultrajson/blob/2e4aba339b0ab08f893590aaa989d12846afc5d6/python/objToJSON.c#L662
    """

    ensure_ascii: bool = True  # limits output to ASCII and escapes all extended characters above 127.
    encode_html_chars: bool = False  # enables special encoding of "unsafe" HTML characters into safer Unicode sequences
    escape_forward_slashes: bool = True  # whether forward slashes (/) are escaped
    indent: int = 0
    sort_keys: bool = False
    allow_nan: bool = True
    reject_bytes: bool = True
    default: ta.Callable[[ta.Any], ta.Any] | None = None  # should return a serializable version of obj or raise TypeError  # noqa
    separators: tuple[str, str] | None = None


##


class UjsonBackend(Backend):
    def dump(self, obj: ta.Any, fp: ta.Any, **kwargs: ta.Any) -> None:
        uj.dump(obj, fp, **kwargs)

    def dumps(self, obj: ta.Any, **kwargs: ta.Any) -> str:
        return uj.dumps(obj, **kwargs)

    def load(self, fp: ta.Any, **kwargs: ta.Any) -> ta.Any:
        return uj.load(fp, **kwargs)

    def loads(self, s: str | bytes | bytearray, **kwargs: ta.Any) -> ta.Any:
        return uj.loads(s, **kwargs)

    def dump_pretty(self, obj: ta.Any, fp: ta.Any, **kwargs: ta.Any) -> None:
        uj.dump(obj, fp, indent=PRETTY_INDENT, **kwargs)

    def dumps_pretty(self, obj: ta.Any, **kwargs: ta.Any) -> str:
        return uj.dumps(obj, indent=PRETTY_INDENT, **kwargs)

    def dump_compact(self, obj: ta.Any, fp: ta.Any, **kwargs: ta.Any) -> None:
        uj.dump(obj, fp, **kwargs)

    def dumps_compact(self, obj: ta.Any, **kwargs: ta.Any) -> str:
        return uj.dumps(obj, **kwargs)


UJSON_BACKEND: UjsonBackend | None
if lang.can_import('ujson'):
    UJSON_BACKEND = UjsonBackend()
else:
    UJSON_BACKEND = None



"""
TODO:
 - inheritance
"""
import dataclasses as dc
import threading
import typing as ta

from .. import collections as col
from .. import lang


class RegistryItem(lang.Abstract):
    pass


RegistryItemT = ta.TypeVar('RegistryItemT', bound=RegistryItem)


@dc.dataclass(frozen=True)
class _KeyRegistryItems:
    key: ta.Any
    items: list[RegistryItem] = dc.field(default_factory=list)
    item_lists_by_ty: dict[type[RegistryItem], list[RegistryItem]] = dc.field(default_factory=dict)

    def add(self, *items: RegistryItem) -> None:
        for i in items:
            self.items.append(i)
            self.item_lists_by_ty.setdefault(type(i), []).append(i)


class Registry:
    def __init__(self) -> None:
        super().__init__()
        self._mtx = threading.Lock()
        self._idct: ta.MutableMapping[ta.Any, _KeyRegistryItems] = col.IdentityKeyDict()
        self._dct: dict[ta.Any, _KeyRegistryItems] = {}

    def register(
            self,
            key: ta.Any,
            *items: RegistryItem,
            identity: bool = False,
    ) -> 'Registry':
        with self._mtx:
            dct: ta.Any = self._idct if identity else self._dct
            if (sr := dct.get(key)) is None:
                sr = dct[key] = _KeyRegistryItems(key)
            sr.add(*items)
        return self

    def get(
            self,
            key: ta.Any,
            *,
            identity: bool | None = None,
    ) -> ta.Sequence[RegistryItem]:
        if identity is None:
            return (
                *self.get(key, identity=True),
                *self.get(key, identity=False),
            )
        dct: ta.Any = self._idct if identity else self._dct
        try:
            return dct[key].items
        except KeyError:
            return ()

    def get_of(
            self,
            key: ta.Any,
            item_ty: type[RegistryItemT],
            *,
            identity: bool | None = None,
    ) -> ta.Sequence[RegistryItemT]:
        if identity is None:
            return (
                *self.get_of(key, item_ty, identity=True),
                *self.get_of(key, item_ty, identity=False),
            )
        dct: ta.Any = self._idct if identity else self._dct
        try:
            sr = dct[key]
        except KeyError:
            return ()
        return sr.item_lists_by_ty.get(item_ty, ())



import dataclasses as dc
import textwrap
import types
import typing as ta

from ... import check
from ... import lang


T = ta.TypeVar('T')
Namespace: ta.TypeAlias = ta.MutableMapping[str, ta.Any]


def create_fn(
        name: str,
        args: ta.Sequence[str],
        body: ta.Sequence[str],
        *,
        globals: Namespace | None = None,  # noqa
        locals: Namespace | None = None,  # noqa
        return_type: lang.Maybe[ta.Any] = lang.empty(),
) -> ta.Callable:
    check.not_isinstance(args, str)
    check.not_isinstance(body, str)

    if locals is None:
        locals = {}  # noqa
    return_annotation = ''
    if return_type.present:
        locals['__dataclass_return_type__'] = return_type()
        return_annotation = '->__dataclass_return_type__'
    args = ','.join(args)
    body = '\n'.join(f'  {b}' for b in body)

    txt = f' def {name}({args}){return_annotation}:\n{body}'

    local_vars = ', '.join(locals.keys())
    txt = f'def __create_fn__({local_vars}):\n{txt}\n return {name}'
    ns: dict[str, ta.Any] = {}
    exec(txt, globals, ns)  # type: ignore
    return ns['__create_fn__'](**locals)


# TODO: https://github.com/python/cpython/commit/8945b7ff55b87d11c747af2dad0e3e4d631e62d6
class FuncBuilder:
    def __init__(self, globals: Namespace) -> None:  # noqa
        super().__init__()

        self.names: list[str] = []
        self.src: list[str] = []
        self.globals = globals
        self.locals: Namespace = {}
        self.overwrite_errors: dict[str, bool] = {}
        self.unconditional_adds: dict[str, bool] = {}

    def add_fn(
            self,
            name: str,
            args: ta.Sequence[str],
            body: ta.Sequence[str],
            *,
            locals: Namespace | None = None,  # noqa
            return_type: lang.Maybe[ta.Any] = lang.empty(),
            overwrite_error: bool = False,
            unconditional_add: bool = False,
            decorator: str | None = None,
    ) -> None:
        if locals is not None:
            self.locals.update(locals)

        # Keep track if this method is allowed to be overwritten if it already exists in the class.  The error is
        # method-specific, so keep it with the name.  We'll use this when we generate all of the functions in the
        # add_fns_to_class call.  overwrite_error is either True, in which case we'll raise an error, or it's a string,
        # in which case we'll raise an error and append this string.
        if overwrite_error:
            self.overwrite_errors[name] = overwrite_error

        # Should this function always overwrite anything that's already in the class?  The default is to not overwrite a
        # function that already exists.
        if unconditional_add:
            self.unconditional_adds[name] = True

        self.names.append(name)

        if return_type.present:
            self.locals[f'__dataclass_{name}_return_type__'] = return_type()
            return_annotation = f' -> __dataclass_{name}_return_type__'
        else:
            return_annotation = ''
        args = ', '.join(args)
        body = textwrap.indent('\n'.join(body), '    ')

        # Compute the text of the entire function, add it to the text we're generating.
        deco_str = '  {decorator}\n' if decorator else ''
        self.src.append(f'{deco_str}  def {name}({args}){return_annotation}:\n{body}')

    def add_fns_to_class(self, cls: type) -> None:
        # The source to all of the functions we're generating.
        fns_src = '\n'.join(self.src)

        # The locals they use.
        local_vars = ','.join(self.locals)

        # The names of all of the functions, used for the return value of the outer function.  Need to handle the
        # 0-tuple specially.
        if not self.names:
            return_names = '()'
        else:
            return_names = f'({",".join(self.names)},)'

        # txt is the entire function we're going to execute, including the bodies of the functions we're defining.
        # Here's a greatly simplified version:
        # def __create_fn__():
        #   def __init__(self, x, y):
        #     self.x = x
        #     self.y = y
        #   @recursive_repr
        #   def __repr__(self):
        #     return f'cls(x={self.x!r},y={self.y!r})'
        # return __init__,__repr__

        txt = f'def __create_fn__({local_vars}):\n{fns_src}\n  return {return_names}'
        ns: dict[str, ta.Any] = {}
        exec(txt, self.globals, ns)  # type: ignore
        fns = ns['__create_fn__'](**self.locals)

        # Now that we've generated the functions, assign them into cls.
        for name, fn in zip(self.names, fns):
            fn.__qualname__ = f'{cls.__qualname__}.{fn.__name__}'
            if self.unconditional_adds.get(name, False):
                setattr(cls, name, fn)
            else:
                already_exists = set_new_attribute(cls, name, fn)

                # See if it's an error to overwrite this particular function.
                if already_exists and (msg_extra := self.overwrite_errors.get(name)):
                    error_msg = (f'Cannot overwrite attribute {fn.__name__} in class {cls.__name__}')
                    if msg_extra:
                        error_msg = f'{error_msg} {msg_extra}'

                    raise TypeError(error_msg)


def set_qualname(cls: type, value: T) -> T:
    if isinstance(value, types.FunctionType):
        value.__qualname__ = f'{cls.__qualname__}.{value.__name__}'
    return value


def set_new_attribute(cls: type, name: str, value: ta.Any) -> bool:
    if name in cls.__dict__:
        return True
    set_qualname(cls, value)
    setattr(cls, name, value)
    return False


def tuple_str(obj_name: str, fields: ta.Iterable[dc.Field]) -> str:
    # Return a string representing each field of obj_name as a tuple member.  So, if fields is ['x', 'y'] and obj_name
    # is "self", return "(self.x,self.y)".

    # Special case for the 0-tuple.
    if not fields:
        return '()'

    # Note the trailing comma, needed if this turns out to be a 1-tuple.
    return f'({",".join([f"{obj_name}.{f.name}" for f in fields])},)'
